{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE TypeFamilies #-}

module RegistryModel where

import Control.Concurrent
import Control.Exception(try,ErrorCall)
import Test.QuickCheck
import Test.QuickCheck.Monadic

import StateModel
import Registry


data RegState = RegState{
    tids :: [Step],
    regs :: [(String,Step)],
    dead :: [Step]
  }
  deriving Show

instance StateModel RegState where
  data Action RegState = Spawn
                       | WhereIs String
                       | Register String Step
                       | Unregister String
		       | Kill Step
    deriving Show

  data Ret RegState = Tid ThreadId
                    | None ()
                    | MaybeTid (Maybe ThreadId)
                    | Caught (Either ErrorCall ())
    deriving (Eq,Show)

  type ActionMonad RegState = IO

  arbitraryAction :: RegState -> Gen (Action RegState)
  arbitraryAction s =
    -- frequency is like oneof, but chooses between alternatives with
    -- integer weights. Adjust these weights to change the
    -- distribution of action types.
    frequency [
           (1,return Spawn),
           (1,Register
                <$> arbitraryName
                <*> elements (tids s)),
           (1,Unregister
                <$> arbitraryName),
           (1,WhereIs
                <$> arbitraryName),
	   (1,Kill
	        <$> elements (tids s))
          ]

  initialState :: RegState
  initialState = RegState [] [] []

  nextState :: RegState -> Action RegState -> Step -> RegState
  nextState s Spawn step =
    s{tids = step:tids s}
  nextState s (Register name tid) _step =
    if positive s (Register name tid)
    then s{regs = (name,tid):regs s}
    else s
  nextState s (Unregister name) _step =
    s{regs = filter ((/=name) . fst) (regs s)}
  nextState s (Kill tid) _ =
    s{dead = tid : dead s,
      regs = filter ((/=tid).snd) (regs s)}
  nextState s _  _ = s

  precondition :: RegState -> Action RegState -> Bool
  precondition s (Register _name step) =
    step `elem` tids s
  precondition s (Kill step) =
    step `elem` tids s
  precondition _ _                    = True

  postcondition ::
    RegState -> Action RegState -> (Step -> Ret RegState) -> Ret RegState -> Bool
  postcondition  s (WhereIs name)       stepValue (MaybeTid mtid) =
    (stepValue <$> lookup name (regs s)) == (Tid <$> mtid)
  postcondition _s Spawn                _         (Tid _)  = True
  postcondition  s (Register name step) _         (Caught res) =
    positive s (Register name step) == (res == Right ())
  postcondition s (Unregister name)   _         (Caught res) =
    positive s (Unregister name) == (res == Right())
  postcondition s (Kill step)           _         (None ()) = True
  postcondition  _ _                    _         _        = False

  monitoring ::
       (RegState, RegState) -> Action RegState -> (Step -> Ret RegState)
    -> Ret RegState -> Property -> Property
  monitoring (s, s') act _ _ =
    classify True lab .
    tabulate "Outcomes" [lab] .
    coverTable "Outcomes" [(lab,4)] .
    tabulate "Registry size" [show (length (regs s))]
    where
      lab = case act of
       Register name step ->
        let problems = ["name in use" | name `elem` map fst (regs s)] ++
	    	       ["pid in use" | step `elem` map snd (regs s)] ++
		       ["pid dead" | step `elem` dead s] in
            "Register: " ++ case problems of
	      [] -> "ok"
	      (x:xs) -> x ++ concatMap (", "++) xs
       _ -> actionName act ++ ": " ++ if positive s act then "ok" else "failed"

  needs :: Action RegState -> [Step]
  needs (Register _ step) = [step]
  needs (Kill step)       = [step]
  needs _                 = []

  perform :: Action RegState -> [Ret RegState] -> IO (Ret RegState)
  perform Spawn _
    = Tid  <$> forkIO (threadDelay 10000000)
  perform (Register name _step) [Tid tid]
    = Caught <$> try (register name tid)
  perform (Unregister name) []
    = Caught <$> try (unregister name)
  perform (WhereIs name) []
    = MaybeTid <$> whereis name
  perform (Kill step) [Tid tid]
    = None <$> kill tid

positive :: RegState -> Action RegState -> Bool
positive s (Register name step) =
     name `notElem` map fst (regs s)
  && step `notElem` map snd (regs s)
  && step `notElem` dead s
positive s (Unregister name) =
     name `elem` map fst (regs s)
positive _s _ = True

arbitraryName :: Gen String
arbitraryName = elements allNames

allNames :: [String]
allNames = ["a", "b", "c", "d", "e"]

prop_Registry :: Script RegState -> Property
prop_Registry s = monadicIO $ do
  _ <- run cleanUp
  _ <- runScript s
  assert True

cleanUp :: IO [Either ErrorCall ()]
cleanUp = sequence
  [try (unregister name) :: IO (Either ErrorCall ())
   | name <- allNames]

kill :: ThreadId -> IO ()
kill tid = do
  killThread tid
  yield
  yield
  
